
# moodss: a Modular Object Oriented Dynamic SpreadSheet
# version 1.0, Copyright (C) 1997 Jean-Luc Fontaine
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# mailto:jfontain@mygale.org

set rcsId {$Id: datatab.tcl,v 1.7 1997/09/21 19:04:34 jfontain Exp $}

class dataTable {                       ;# module data view in the form of a table, that can be sorted using a column as a reference
    proc dataTable {this parentPath args} composite {[new scroll table $parentPath] $args} {
        set path $composite::($composite::($this,base),scrolled,path)
        # table data is at global level
        $path configure -rows 2 -state disabled -titlerows 1 -roworigin -1 -colstretchmode unset -variable ${this}data
        $path tag configure title -font $widget::(default,ButtonFont) -anchor c
        $path tag configure highlight -background [$path tag cget sel -background]      ;# use selection background for highlighting

        bindtags $path "$path . all"           ;# remove all class bindings for we do not use any and they would cause interferences
        bind $path <ButtonRelease-1> "dataTable::columnSort $this %x %y"

        set dataTable::($this,sortOrder) increasing
        set dataTable::($this,tablePath) $path
        composite::complete $this
    }

    proc ~dataTable {this} {
        global ${this}data
        catch {unset ${this}data}                                                                      ;# eventually free table data
    }

    proc options {this} {
        return [list\
            [list -columns columns Columns 0 0]\
            [list -titlefont titleFont TitleFont $widget::(default,ButtonFont) $widget::(default,ButtonFont)]\
            [list -moduledata moduleData ModuleData {} {}]\
        ]
    }

    proc set-columns {this value} {
        $dataTable::($this,tablePath) configure -cols $value
    }

    proc set-titlefont {this value} {
       $dataTable::($this,tablePath) tag configure title -font $value
    }

    proc set-moduledata {this value} {
        global ${this}data
        upvar $value data                                                                            ;# parameter is data array name

        set list {}
        for {set column 0} {$column<$data(columns)} {incr column} {
            lappend list -1,$column $data($column,label)
        }
        array set ${this}data $list                                                                              ;# set table titles
        $dataTable::($this,tablePath) configure -cols $column                                        ;# and fit to number of columns
        highlightColumn $this [lindex $data(sort) 0]
        set dataTable::($this,sortOrder) [lindex $data(sort) 1]
        trace variable data(updates) w "dataTable::update $this"                                        ;# track module data updates
    }

    proc highlightColumn {this {column {}}} {                             ;# highlight column or return currently highlighted column
        set path $dataTable::($this,tablePath)
        set highlighted [lindex [$path tag col highlight] 0]
        if {[string length $column]==0} {                                     ;# return currently highlighted column if no arguments
            return $highlighted
        } else {
            if {[string length $highlighted]>0} {
                $path tag col {} $highlighted                                 ;# first eventually clear currently highlighted column
            }
            $path tag col highlight $column
        }
    }

    proc columnSort {this x y} {                           ;# sort table rows using the column that the user selected as a reference
        set path $dataTable::($this,tablePath)
        set number [expr {[$path cget -rows]-1}]                                                    ;# calculate number of data rows
        if {$number==0} return
        scan [$path index @$x,$y] %d,%d row column
        if {$row>=0} return                                                                                     ;# not the title row

        if {[string compare $column [highlightColumn $this]]==0} {                                           ;# sort the same column
            if {[string compare $dataTable::($this,sortOrder) increasing]==0} {                                ;# but toggle sorting
                set dataTable::($this,sortOrder) decreasing
            } else {
                set dataTable::($this,sortOrder) increasing
            }
        } else {                                                                ;# sort for the first time or for a different column
            set dataTable::($this,sortOrder) increasing
        }
        highlightColumn $this $column                                                                     ;# highlight sorted column
        update $this                                                                                     ;# update table immediately
    }

    proc update {this args} {                                   ;# update display using module data. ignore eventual trace arguments
        if {[string length $composite::($this,-moduledata)]==0} return                                   ;# nothing to do if no data

        global ${this}data
        upvar $composite::($this,-moduledata) data

        set path $dataTable::($this,tablePath)

        set cursor [$path cget -cursor]                                                                               ;# save cursor
        $path configure -cursor watch                                                                  ;# show user that we are busy
        ::update idletasks

        set sortColumn [highlightColumn $this]                                                 ;# use highlighted column for sorting
        set lists {}
        foreach name [array names data *,0] {
            scan $name %u row
            lappend lists [list $row $data($row,$sortColumn)]
        }

        set lists [lsort -$dataTable::($this,sortOrder) -$data($sortColumn,type) -index 1 $lists]
        set row 0
        foreach pair $lists {
            set index [lindex $pair 0]
            for {set column 0} {$column<$data(columns)} {incr column} {
                set ${this}data($row,$column) $data($index,$column)
            }
            incr row
        }
        $path configure -rows [incr row]                                                ;# fit to data (take into account title row)
        highlightColumn $this $sortColumn                                      ;# refresh highlighted column in case rows were added

        $path configure -cursor $cursor                                                                            ;# restore cursor
        ::update idletasks
    }
}
